#!/usr/bin/env perl

package acid;

use strict;
use POSIX 'strftime';
use IO::Socket::INET;
use lib::Module::Refresh;
use Data::Dumper;

our $VERSION = '3.0.6';

$SIG{HUP} = \&sig_hup;
$SIG{INT} = \&sig_int;

our ($eob, $uplink) = 0;
my ($config, $clientcount, $module_queue) = 'acid.conf';
my (%conf, %lc, %client, %uid, %chan, %server, %map);
my (%module, %prefix);
my ($socket, $refresher);

%prefix = ('~' => 1, '&' => 2, '@' => 3, '%' => 4, '+' => 5, ''  => 6);

printf "acidictive-%s started\n", $VERSION;

my $error = config_load();
die "$config: $error" if $error;

if ($conf{livenet}) {
	my $pid = fork;
	die "Failed to fork: $!" if !defined($pid);
	exit if $pid;

	open STDIN, '>', '/dev/null';
	open STDOUT, '>', '/dev/null';
	open STDERR, '>', '/dev/null';
}

$refresher = Module::Refresh->new;

for (split(/ |, ?/, $conf{modules_startup})) {
	my $error = module_load($_);
	die "Failed to load startup module $_: $error" if $error;
}

$socket = IO::Socket::INET->new(
	PeerAddr		=> $conf{server},
	PeerPort		=> $conf{port},
	LocalAddr	=> $conf{localaddr},
	Proto			=> 'tcp'
) || die "Failed to connect to $conf{server}:$conf{port}: $!";

$server{$conf{sid}} = $conf{name};

while (my @raw = split (' ', <$socket>)) {
	#map s/\r\n//g, @raw;
	local $/ = "\r\n";
	chomp @raw;
	local $/ = "\n";

	debug("<- @raw");

	if ($module_queue) {
		module_unload($module_queue);
		$module_queue = '';
	}

	# 439
	# :irc.core.ws 439 * :Please wait while we process your connection.
	if ($raw[1] eq '439') {
		raw(undef, 'PASS %s TS 6 :%s', $conf{sendpass}, $conf{sid});
		raw(undef, 'CAPAB :%s', $conf{capab});
		raw(undef, 'SERVER %s 1 :(H) %s', $conf{name}, $conf{desc});
		raw(undef, 'SVINFO 6 6 0 :%d', time);
	}

	# PASS
	# PASS r00t!S3RV TS 6 0XY
	elsif ($raw[0] eq 'PASS') {
		my ($pass, $tsversion, $sid) = ($raw[1], @raw[3..4]);

		close $socket if $conf{acceptpass} && $pass ne $conf{acceptpass};

		$uplink = $sid;
		push @{$map{$sid}}, $conf{sid};

		callhooks('on_pass', $pass, $tsversion, $sid);
	}

	# SERVER
	# SERVER irc.core.ws 1 :Rizon IRC Network - Client Server
	elsif ($raw[0] eq 'SERVER') {
		my ($server, $hops, $desc) = (@raw[1..2], substr(join(' ', @raw[3..$#raw]), 1));

		$server{$uplink} = $server;

		callhooks('on_server', $server, $hops, $desc);
	}

	# PING
	# PING :0XY
	elsif ($raw[0] eq 'PING') {
		my $sid = substr($raw[1], 1);

		raw(undef, 'PONG :%s', $sid);

		callhooks('on_ping', $sid);
	}

	# SQUIT
	# SQUIT 069 :bye bye
	elsif ($raw[0] eq 'SQUIT') {
		my ($sid, $msg) = ($raw[1], substr(join(' ', @raw[2..$#raw]), 1));

		callhooks('on_squit', $sid, $msg);
	}

	# SID
	# :0XY SID nessun-is-a-pain.hub 2 42T :(H) FOR GREAT JUSTICE!!!
	elsif ($raw[1] eq 'SID') {
		my ($introsid, $server, $hops, $sid, $desc) = (substr($raw[0], 1), @raw[2..4], substr(join(' ', @raw[5..$#raw]), 1));

		$server{$sid} = $server;
		push @{$map{$introsid}}, $sid;

		callhooks('on_sid', $introsid, $server, $hops, $sid, $desc);
	}

	# UID
	# :0XY UID Blake 1 1270327294 +Sirx ~blake blake.blake 98.192.115.215 0XYAAAABC 1270327294 c-98-192-115-215.hsd1.ga.comcast.net :yourname
	elsif ($raw[1] eq 'UID') {
		my ($sid, $nick, $hops, $ts, $umode, $ident, $cloak, $ip, $uid, $host, $gecos) = (substr($raw[0], 1), @raw[2..9], $raw[11], substr(join(' ', @raw[12..$#raw]), 1));

		callhooks('on_uid', $sid, $nick, $hops, $ts, $umode, $ident, $cloak, $ip, $uid, $host, $gecos);
	}

	# EOB
	# :0XY EOB
	elsif ($raw[1] eq 'EOB') {
		my $sid = substr($raw[0], 1);

		callhooks('on_eob', $sid);

		if ($sid eq $uplink) {
			raw($conf{sid}, 'EOB');
			$eob = 1;
			
			for (split(/ |, ?/, $conf{modules})) {
				my $error = module_load($_);
				debug("OOPS: failed to load module $_: $error") if $error;
			}
		}
	}

	# SJOIN
	# :0XY SJOIN 1270009019 #holz +nstz :0XYAAAABC @0XYAAAAAI 0XYAAAAAN 42TAAAAAJ
	elsif ($raw[1] eq 'SJOIN') {
		my ($sid, $ts, $chan, $modes, @users) = (substr($raw[0], 1), @raw[2..$#raw]);

		$modes .= ' '.shift(@users) if $modes =~ /l/;
		$modes .= ' '.shift(@users) if $modes =~ /k/;

		if ($users[0] eq ':') {
			@users = ();
		}
		else {
			$users[0] = substr($users[0], 1);
		}

		tssync($chan, $ts) if $chan{$chan};

		callhooks('on_sjoin', $sid, $chan, $ts, $modes, @users);
	}

	# TMODE
	# :0XYAAAAAI TMODE 1270009025 #lisa +c-vN+o 0XYAAAAAD 0XYAAAAAD
	elsif ($raw[1] eq 'TMODE') {
		my ($id, $ts, $chan, $modes) = (substr($raw[0], 1), @raw[2..3], join(' ', @raw[4..$#raw]));

		callhooks('on_tmode', $id, $ts, $chan, $modes);
	}

	# BMASK
	# :0XY BMASK 1270009025 #lisa b :Kokizzzle!*@* Kokizzzle!*@rizon.staff Kokizzzle!*Kokizzzle@rizon.staff
	elsif ($raw[1] eq 'BMASK') {
		my ($sid, $ts, $chan, $type, @masks) = (substr($raw[0], 1), @raw[2..$#raw]);
		$masks[0] = substr($masks[0], 1);

		if ($type eq 'b') {
			$type = 'ban';
		}
		elsif ($type eq 'e') {
			$type = 'except';
		}
		elsif ($type eq 'I') {
			$type = 'invex';
		}

		my $time = time;
		for (@masks) {
			next if $chan{$chan}{$type}{$_};
			$chan{$chan}{$type}{$_}{ts} = $time;
			$chan{$chan}{$type}{$_}{setby} = $server{$sid};
		}
	}

	# TBURST
	# :irc.core.ws TBURST 1270367145 #k 1270372774 Holz!Holz@asd.asdad :hi
	elsif ($raw[1] eq 'TBURST') {
		my ($sid, $ts, $chan, $topicts, $topicby, $topic) = (getid(substr($raw[0], 1)), @raw[2..5], substr(join(' ', @raw[6..$#raw]), 1));

		$chan{$chan}{topicts} = $topicts;
		$chan{$chan}{topicby} = $topicby;
		$chan{$chan}{topic} = $topic;
	}

	# TOPIC
	# :0XYAAAAAI TOPIC #k :hi penis
	elsif ($raw[1] eq 'TOPIC') {
		my ($uid, $chan, $topic) = (substr($raw[0], 1), $raw[2], substr(join(' ', @raw[3..$#raw]), 1));

		if ($topic) {
			$chan{$chan}{topicts} = time;
			$chan{$chan}{topicby} = "$client{$uid}{nick}!$client{$uid}{ident}\@$client{$uid}{cloak}";
			$chan{$chan}{topic} = $topic;
		}
		else {
			delete $chan{$chan}{topicts};
			delete $chan{$chan}{topicby};
			delete $chan{$chan}{topic};
		}

		callhooks('on_topic', $uid, $chan, $topic);
	}

	# JOIN
	# :0XYAAAABD JOIN 1270367145 #k +
	elsif ($raw[1] eq 'JOIN') {
		my ($uid, $ts, $chan, $mode) = (substr($raw[0], 1), @raw[2..3], join(' ', @raw[4..$#raw]));

		if (!$client{$uid}) {
			debug("OOPS: JOIN $chan on non-existent client: $uid");
			next;
		}

		$chan{$chan}{users}{$uid} = '';
		$client{$uid}{chans}{$chan} = '';

		tssync($chan, $ts);

		callhooks('on_join', $uid, $ts, $chan, $mode);
	}

	# MODE
	# :0XYAAAAAI MODE 0XYAAAAAI :-w
	elsif ($raw[1] eq 'MODE') {
		my ($uid, $target, $modes) = (substr($raw[0], 1), getid($raw[2]), substr($raw[3], 1));

		if (!$client{$target}) {
			debug("OOPS MODE $uid $modes on non-existent client: $target");
			next;
		}

		callhooks('on_mode', $uid, $target, $modes);
	}

	# SMODE
	# :4SS SMODE 4SSAAAAAC :+W
	elsif ($raw[1] eq 'SMODE') {
		my ($sid, $target, $modes) = (substr($raw[0], 1), getid($raw[2]), substr($raw[3], 1));

		if (!$client{$target}) {
			debug("OOPS SMODE $sid $modes on non-existent client: $target");
			next;
		}

		callhooks('on_mode', $sid, $target, $modes);
	}

	# PART
	# :0XYAAAAAI PART #k
	# :0XYAAAAAI PART #k :bye
	elsif ($raw[1] eq 'PART') {
		my ($uid, $chan, $msg) = (substr($raw[0], 1), $raw[2], substr(join(' ', @raw[3..$#raw]), 1));

		if (!$client{$uid}) {
			debug("OOPS: PART $chan on non-existent client: $uid");
			next;
		}
		
		callhooks('on_part', $uid, $chan, $msg);
	}

	# KICK
	# :0XYAAAAAI KICK #k 0XYAAAAAI :gtfo
	elsif ($raw[1] eq 'KICK') {
		my ($uid, $chan, $targetuid, $msg) = (substr($raw[0], 1), @raw[2..3], substr(join(' ', @raw[4..$#raw]), 1));

		if (!$client{$targetuid}) {
			debug("OOPS: KICK $uid $chan $targetuid on non-existent client: $targetuid");
			next;
		}

		callhooks('on_kick', $uid, $chan, $targetuid, $msg);
	}

	# NICK
	# :0XYAAAAAI NICK Lisa :1270373060
	elsif ($raw[1] eq 'NICK') {
		my ($uid, $newnick, $ts) = (substr($raw[0], 1), $raw[2], substr($raw[3], 1));

		if (!$client{$uid}) {
			debug("OOPS: NICK $newnick on non-existent client: $uid");
			next;
		}

		callhooks('on_nick', $uid, $newnick, $ts);
	}

	# QUIT
	# :0XYAAAABD QUIT :bye
	elsif ($raw[1] eq 'QUIT') {
		my ($uid, $msg) = (substr($raw[0], 1), substr(join(' ', @raw[2..$#raw]), 1));


		if (!$client{$uid}) {
			debug("OOPS: QUIT $uid on non-existent client: $uid");
			next;
		}

		callhooks('on_quit', $uid, $msg);
	}

	# KILL
	# :02SAAAAFO KILL 0XYAAAAAN :NickServ (GHOST command used by Holz)
	elsif ($raw[1] eq 'KILL') {
		my ($uid, $victim, $msg) = (substr($raw[0], 1), $raw[2], substr(join(' ', @raw[3..$#raw]), 1));
		$msg = "Killed ($msg)";

		if (!$client{$victim}) {
			debug("OOPS: KILL $uid on non-existent client: $victim");
			next;
		}

		callhooks('on_quit', $victim, $msg);
	}

	# PRIVMSG
	# :0XYAAAAAI PRIVMSG #lisa :asdf asdf
	elsif ($raw[1] eq 'PRIVMSG') {
		my ($uid, $target, $msg) = (getid(substr($raw[0], 1)), $raw[2], substr(join(' ', @raw[3..$#raw]), 1));
		$target = getid($target) if substr($target, 0, 1) ne '#';

		$client{$uid}{idle} = time;

		if (substr($msg, 0, 1) eq chr(1) && substr($msg, -1) eq chr(1)) {
			my $ctcp = substr($msg, 1, -1);
			next if uc(substr($ctcp, 0, 6)) eq 'ACTION';

			if (substr($target, 0, 1) ne '#') {
				if (uc($ctcp) eq 'VERSION') {
					notice($target, $uid, "\001%s acidictive-%s\001", $ctcp, $VERSION);
				}
				elsif (uc($ctcp) eq 'TIME') {
					notice($target, $uid, "\001%s %s\001", $ctcp, scalar(localtime));
				}
			}

			callhooks('on_ctcp', $uid, $target, $ctcp);
		}
		else {
			callhooks('on_privmsg', $uid, $target, $msg);
		}
	}

	# NOTICE
	# :0XYAAAAAI NOTICE RootServ :hello
	elsif ($raw[1] eq 'NOTICE') {
		my ($id, $target, $msg) = (substr($raw[0], 1), $raw[2], substr(join(' ', @raw[3..$#raw]), 1));
		next if !$eob;
		$id = getid($id);
		$target = getid($target) if substr($target, 0, 1) ne '#';

		# WHOIS
		# :0XY NOTICE 1ABAAAAAA :*** Notice -- Holz (Holz@Holz) is doing a whois on you [irc.core.ws]
		if ($server{$id} && "@raw[8..13]" eq 'is doing a whois on you') {
			callhooks('on_whois', $id, $raw[6], $target);
		}

		callhooks('on_notice', $id, $target, $msg);
	}

	# TIME
	# :0XYAAAAAI TIME :acid.rizon.net
	elsif ($raw[1] eq 'TIME') {
		my ($id, $target) = (getid(substr($raw[0], 1)), substr($raw[2], 1));

		raw($conf{sid}, '391 %s %s :%s', $id, $conf{name}, scalar(localtime));

		callhooks('on_time', $id, $target);
	}

	# WHOIS
	# :0XYAAAAAI WHOIS acid.rizon.net :RootServ
	elsif ($raw[1] eq 'WHOIS') {
		my ($id, $me, $target) = (getid(substr($raw[0], 1)), $raw[2], substr($raw[3], 1));
		
		if ($lc{lc($target)} && $client{getid($target)}{sid} eq $conf{sid}) {
			my $uid = getid($target);
			my $nick = $client{$uid}{nick};

			raw($conf{sid}, '311 %s %s %s %s * :%s', $id, $nick, $client{$uid}{ident}, $client{$uid}{cloak}, $client{$uid}{gecos});

			my $chans;
			for (sort { lc($a) cmp lc($b) } keys(%{$client{$uid}{chans}})) {
				if ($client{$id}{umode}{a} || defined($chan{$_}{users}{$id}) || (!$chan{$_}{mode}{s} && (!$client{$uid}{umode}{p} || $client{$id}{umode}{o}))) {
					$chans .= "$client{$uid}{chans}{$_}$_ ";
				}
			}
			if ($chans) {
				raw($conf{sid}, '319 %s %s :%s', $id, $nick, substr($chans, 0, -1));
			}

			if ($client{$id}{umode}{o}) {
				raw($conf{sid}, '312 %s %s %s :%s', $id, $nick, $conf{name}, $conf{desc});
			}
			else {
				raw($conf{sid}, '312 %s %s *.rizon.net :Where are you?', $id, $nick);
			}

			if ($client{$uid}{umode}{G} && !$client{$uid}{umode}{g}) {
				raw($conf{sid}, '716 %s %s :is in +G mode (server side ignore with the exception of common channels)', $id, $nick);
			}

			if ($client{$uid}{umode}{g}) {
				raw($conf{sid}, '716 %s %s :is in +g mode (server side ignore)', $id, $nick);
			}

			if ($client{$uid}{umode}{o}) {
				my $oper = 'an IRC Operator';
				$oper .= ' - Network Administrator' if $client{$uid}{umode}{N};
				$oper .= ' - Server Administrator' if $client{$uid}{umode}{a} && !$client{$uid}{umode}{N};
				raw($conf{sid}, '313 %s %s :is %s', $id, $nick, $oper);
			}

			if ($client{$uid}{umode}{r}) {
				raw($conf{sid}, '307 %s %s :has identified for %s', $id, $nick, ($client{$id}{umode}{o} && $client{$uid}{su} ? $client{$uid}{su} : 'this nick'));
			}

			if ($client{$uid}{umode}{S}) {
				raw($conf{sid}, '671 %s %s :is using a secure connection', $id ,$nick);
			}

			if ($client{$id}{umode}{o}) {
				if ($client{$uid}{certfp}) {
					raw($conf{sid}, '276 %s %s :is using a client certificate', $id, $nick);
				}

				raw($conf{sid}, '310 %s %s :is using modes +%s authflags: %s',
					$id, $nick, join('', sort(keys(%{$client{$uid}{umode}}))), ($client{$uid}{authflags} ? $client{$uid}{authflags} : '[none]'));

				raw($conf{sid}, '338 %s %s :is actually %s@%s [%s]',
					$id, $nick, $client{$uid}{ident}, $client{$uid}{host}, ($client{$uid}{authflags} && $client{$uid}{authflags} =~ /=/ ? '255.255.255.255' : $client{$uid}{ip}));
			}

			if (!$client{$uid}{umode}{p} || $client{$id}{umode}{a}) {
				raw($conf{sid}, '317 %s %s %d %d :seconds idle, signon time', $id, $nick, (time - ($client{$uid}{idle} ? $client{$uid}{idle} : $^T)), $^T);
			}

			callhooks('on_whois', $conf{sid}, $id, $uid);
		}
		else {
			raw($conf{sid}, '401 %s %s :No such nick/channel', $id, $target);
		}

		raw($conf{sid}, '318 %s %s :End of /WHOIS list.', $id, $target);
	}

	# STATS
	# :0XYAAAAAI STATS u :acid.rizon.net
	elsif ($raw[1] eq 'STATS') {
		my ($id, $letter, $target) = (getid(substr($raw[0], 1)), $raw[2], substr($raw[3], 1));

		# u
		if ($letter eq 'u') {
			raw($conf{sid}, '242 %s :Server Up %s', $id, duration(time - $^T));
			raw($conf{sid}, '219 %s %s :End of /STATS report', $id, $letter);
		}

		callhooks('on_stats', $id, $letter, $target);
	}

	# VERSION
	# :0XYAAAAAC VERSION :acid.rizon.net
	elsif ($raw[1] eq 'VERSION') {
		my ($id, $target) = (getid(substr($raw[0], 1)), substr($raw[2], 1));

		raw($conf{sid}, '351 %s acidictive-%s %s :MODULES=%s', $id, $VERSION, $conf{name}, join('+', sort(keys(%module))));

		callhooks('on_version', $id, $target);
	}

	# ENCAP
	elsif ($raw[1] eq 'ENCAP') {
		my $sid = getid(substr($raw[0], 1));

		# AUTHFLAGS
		# :irc.core.ws ENCAP * AUTHFLAGS grimlock $=^_><|
		if ($raw[3] eq 'AUTHFLAGS') {
			my ($uid, $authflags) = ($raw[4], $raw[5]);

			if (!$lc{lc($uid)}) {
				debug("OOPS: ENCAP AUTHFLAGS on non-existent client: $uid");
				next;
			}

			$uid = getid($uid);
			$authflags = substr($authflags, 1) if substr($authflags, 0, 1) eq ':';

			$client{$uid}{authflags} = $authflags;

			callhooks('on_encap_authflags', $sid, $uid, $authflags);
		}

		# CERTFP
		# :irc.core.ws ENCAP * CERTFP grimlock :BC78C0AEEF3A82CBA145C03E7403513110F83015
		elsif ($raw[3] eq 'CERTFP') {
			my ($uid, $certfp) = ($raw[4], $raw[5]);

			if (!$lc{lc($uid)}) {
				debug("OOPS: ENCAP CERTFP on non-existent client: $uid");
				next;
			}

			$uid = getid($uid);
			$certfp = substr($certfp, 1) if substr($certfp, 0, 1) eq ':';

			$client{$uid}{certfp} = $certfp;

			callhooks('on_encap_certfp', $sid, $uid, $certfp);
		}

		# CHGHOST
		# :02S ENCAP * CHGHOST Holz :asd.asdad
		elsif ($raw[3] eq 'CHGHOST') {
			my ($uid, $host) = ($raw[4], $raw[5]);

			if (!$lc{lc($uid)}) {
				debug("OOPS: ENCAP CHGHOST on non-existent client: $uid");
				next;
			}

			$uid = getid($uid);
			$host = substr($host, 1) if substr($host, 0, 1) eq ':';

			$client{$uid}{cloak} = $host;

			callhooks('on_encap_chghost', $sid, $uid, $host);
		}

		# CHGIDENT
		# :irc.core.ws ENCAP * CHGIDENT Blake test
		elsif ($raw[3] eq 'CHGIDENT') {
			my ($uid, $ident) = ($raw[4], $raw[5]);

			if (!$lc{lc($uid)}) {
				debug("OOPS: ENCAP CHGIDENT on non-existent client: $uid");
				next;
			}

			$uid = getid($uid);
			$ident = substr($ident, 1) if substr($ident, 0, 1) eq ':';

			$client{$uid}{ident} = $ident;

			callhooks('on_encap_chgident', $sid, $uid, $ident);
		}

		# SVSMODE
		# :02S ENCAP * SVSMODE RootServ 1270061317 :-r
		# :02S ENCAP * SVSMODE Lisa 1270071905 +r+qd :1270071905
		elsif ($raw[3] eq 'SVSMODE') {
			my ($uid, $ts, $modes, $ts2) = ($raw[4], @raw[5..$#raw]);

			if (!$lc{lc($uid)}) {
				debug("OOPS: ENCAP SVSMODE on non-existent client: $uid");
				next;
			}

			$uid = getid($uid);
			$modes = substr($modes, 1) if substr($modes, 0, 1) eq ':';

			callhooks('on_mode', $sid, $uid, $modes);
		}

		# SU
		# :irc.core.ws ENCAP * SU Kokizzzle Kokizzzle
		# :02S ENCAP * SU :adadads
		elsif ($raw[3] eq 'SU') {
			my ($uid, $su) = @raw[4..5];
			$uid = substr($uid, 1) if substr($uid, 0, 1) eq ':';

			if (!$lc{lc($uid)}) {
				debug("OOPS: ENCAP SU on non-existent client: $uid");
				next;
			}

			$uid = getid($uid);
			$su = substr($su, 1) if substr($su, 0, 1) eq ':';

			if ($su) {
				$client{$uid}{su} = $su;
			}
			else {
				delete $client{$uid}{su};
			}

			callhooks('on_encap_su', $sid, $uid, $su);
		}

	}

	callhooks('on_check') if $eob;
}


# raw([$prefix], $cmd, @args)
sub raw {
	my $prefix = shift;
	my $raw = ($prefix ? ":$prefix " : '').sprintf(shift, @_);

	printf $socket "%s\r\n", $raw;
	debug("-> $raw");
}

# debug($info)
sub debug {
	my $info = shift;

	printf "[%s] %s\n", strftime('%H:%M:%S', localtime), $info if !$conf{livenet};

	if ($conf{debuglog}) {
		open LOG, '>>', $conf{debuglog} || return;
		printf LOG "%s %s\n", scalar(localtime), $info;
		close LOG;
	}
}

# config_load()
sub config_load {
	my %test = do($config);

	return $! if $!;

	if ($@) {
		chomp $@;
		return "$@";
	}

	for (qw(server port sendpass sid name desc capab logchan)) {
		return "'$_' is a required config setting!" if !$test{$_};
	}

	%conf = %test;

	return;
}

# config_save()
sub config_save {
	local $Data::Dumper::Sortkeys = 1;
	my @dump = split(/\n/, Dumper(\%conf));

	open CONFIG, '>', $config || return $!;
	printf CONFIG "%s\n", substr($_, 10) for splice(@dump, 1, -1);
	close CONFIG;

	return;
}

# module_load($module)
sub module_load {
	my $module = shift;

	my $test = quotemeta($module).'.pm';
	if (system("perl -c mod/$test")) {
		my $error;
		for (`perl -c mod/$test 2>&1`) {
			y/\t/ /;
			$error .= "\n$_";
		}
		return $error;
	}

	eval { require "mod/$module.pm" };

	if ($@) {
		chomp $@;
		return "$@";
	}

	$module{$module} = $module->new(
		module	=> \%module,
		prefix	=> \%prefix,
		conf		=> \%conf,
		lc			=> \%lc,
		client	=> \%client,
		uid		=> \%uid,
		chan		=> \%chan,
		server	=> \%server,
		#map		=> \%map
	);

	return;
}

# module_prepare_unload($module)
sub module_prepare_unload {
	my $module = shift;

	$module_queue = $module;
}

# module_unload($module)
sub module_unload {
	my $module = shift;

	return 'no such module loaded' if !$module{$module};

	$module{$module}->on_unload;
	$refresher->unload_module("mod/$module.pm");
	delete $module{$module};

	return;
}

# module_refresh($module)
sub module_refresh {
	my $module = shift;

	return 'no such module loaded' if !$module{$module};

	my $test = quotemeta($module).'.pm';
	if (system("perl -c mod/$test")) {
		my $error;
		for (`perl -c mod/$test 2>&1`) {
			y/\t/ /;
			$error .= "\n$_";
		}
		return $error;
	}

	$refresher->refresh_module("mod/$module.pm");

	return;
}

# shutdown_($sender, [$target])
sub shutdown_ {
	my ($sender, $target) = @_;

	$eob = 0;

	notice($sender, $target, 'Shutting down after %s...', duration(time - $^T)) if $target;

	notice($sender, $target, 'Saving config...') if $target;
	my $error = config_save();
	notice($sender, $target, 'Failed to save config, continuing to shut down anyway: %s', $error) if $target && $error;

	notice($sender, $target, 'Bye!') if $target;

	callhooks('on_shutdown');

	for (keys(%client)) {
		quit($_, 'shutting down') if $client{$_}{sid} eq $conf{sid};
	}

	if ($target) {
		raw($conf{sid}, 'SQUIT %s :received SHUTDOWN from %s', $conf{sid}, $target);
	}
	else {
		raw($conf{sid}, 'SQUIT %s :received SIGINT', $conf{sid});
	}
}

# sig_hup()
sub sig_hup {
	notice($conf{sid}, $conf{logchan}, 'Received SIGHUP, rehashing %s.', $config);
	my $error = config_load();
	if ($error) {
		notice($conf{sid}, $conf{logchan}, '%s could not be rehashed: %s', $config, $error);
	}
	else {
		notice($conf{sid}, $conf{logchan}, '%s has been rehashed.', $config);
	}
}

# sig_int()
sub sig_int {
	notice($conf{sid}, $conf{logchan}, 'Received SIGINT, shutting down.', $config);

	shutdown_($conf{sid});
}

# callhooks($sub, @args)
sub callhooks {
	my ($sub, @args) = @_;
	no strict 'refs';

	debug("*** CALLHOOKS (EOB: $eob): $sub(".join(', ', @args).')') if $sub ne 'on_check';

	&$sub(@args) if defined(&$sub);

	#return if !$eob && $sub ne 'on_eob';

	for (values(%module)) {
		$_->$sub(@args) if $_->can($sub);
	}
}




# getid($name)
sub getid {
	my $name = shift;

	return $name if $name =~ /^\d[A-Z\d]{2}(?:[A-Z][A-Z\d]{5})?$/;

	return $uid{$lc{lc($name)}} if $lc{lc($name)};

	for (keys(%server)) {
		return $_ if lc($server{$_}) eq lc($name);
	}

	debug("OOPS: getid on non-existent name: $name");

	return;
}

# getname($id)
sub getname {
	my $id = shift;

	return $id if $id !~ /^\d[A-Z\d]{2}(?:[A-Z][A-Z\d]{5})?$/;

	return $server{$id} if $server{$id};

	return $client{$id}{nick} if $client{$id};

	debug("OOPS: getname on non-existent id: $id");

	return;
}

# duration($secs)
sub duration {
	my $secs = shift;

	return '0secs' if !$secs;

	my @gmt = gmtime($secs);

	return	($gmt[7] ?  $gmt[7]														.'day'.($gmt[7] > 1 ? 's' : '') : '').
				($gmt[2] ? ($gmt[7]							  ? ' ' : '').$gmt[2].'hr' .($gmt[2] > 1 ? 's' : '') : '').
				($gmt[1] ? ($gmt[7] || $gmt[2]				? ' ' : '').$gmt[1].'min'.($gmt[1] > 1 ? 's' : '') : '').
				($gmt[0] ? ($gmt[7] || $gmt[2] || $gmt[1] ? ' ' : '').$gmt[0].'sec'.($gmt[0] > 1 ? 's' : '') : '');
}

# getinfo($uid)
sub getinfo {
	my $uid = shift;

	###
	return 'no uid' if !$uid;
	return "unknown uid: $uid" if !$client{$uid};
	###

	my $nick = $client{$uid}{nick};
	my $ident = $client{$uid}{ident};
	my $host = $client{$uid}{host};
	my $ip = $client{$uid}{ip};
	my $cloak = $client{$uid}{cloak};
	my $gecos = $client{$uid}{gecos};
	my $umode = join('', sort(keys(%{$client{$uid}{umode}})));
	my $server = $server{$client{$uid}{sid}};

	return "$nick!$ident\@$host:$ip:$cloak($gecos)(+$umode)[$server]";
}


# setupclients(%clients)
sub setupclients {
	my %clients = @_;

	for my $c (keys(%clients)) {
		uid($clients{$c}{nick}, $clients{$c}{ident}, $clients{$c}{host}, $clients{$c}{ip},
			$clients{$c}{cloak}, $clients{$c}{gecos}, $clients{$c}{umode}, $clients{$c}{authflags}
		);

		my $uid = getid($clients{$c}{nick});

		if ($clients{$c}{nickserv}) {
			privmsg($uid, 'NickServ@services.rizon.net', 'IDENTIFY %s', $clients{$c}{nickserv});
		}

		if ($clients{$c}{chans}) {
			for (keys(%{$clients{$c}{chans}})) {
				join_($uid, $_, '+nt', $clients{$c}{chans}{$_});
			}
		}

	}	
}
# removeclients(%clients)
sub removeclients {
	my %clients = @_;

	for (keys(%clients)) {
		quit(getid($clients{$_}{nick}), 'module unloaded');
	}
}


# tssync($chan, $ts)
sub tssync {
	my ($chan, $ts) = @_;
	my $myts = $chan{$lc{lc($chan)}}{ts};

	if ($myts > $ts) {
		$chan{$chan}{ts} = $ts;

		for (keys(%{$chan{$chan}{mode}})) {
			delete $chan{$chan}{mode}{$_} if $_ ne 'z';
		}

		for (keys(%{$chan{$chan}{users}})) {
			$chan{$chan}{users}{$_} = '';
			$client{$_}{chans}{$chan} = '';
		}
	}

	#elsif ($myts < $ts) {
	#}
}


# uid($nick, $ident, $host, $ip, $cloak, $gecos, $umode, [$authflags])
sub uid {
	my ($nick, $ident, $host, $ip, $cloak, $gecos, $umode, $authflags) = @_;
	my ($count, $uid) = $clientcount;

	my @chars = ('A'..'Z', 0..9);
	while ($count) {
		$uid = $chars[$count % 36].$uid;
		$count = int($count / 36);
	}

	$uid = $conf{sid}.'A'x(6 - length($uid)).$uid;

	raw($conf{sid}, 'UID %s 1 %d %s %s %s %s %s 0 %s :%s', $nick, time, $umode, $ident, $cloak, $ip, $uid, $host, $gecos);

	callhooks('on_uid', $conf{sid}, $nick, 0, time, $umode, $ident, $cloak, $ip, $uid, $host, $gecos);

	authflags($nick, $authflags) if $authflags;

	$clientcount++;
}

# privmsg($sender, $target, $msg, @args)
sub privmsg {
	my $sender = shift;
	my $target = shift;
	my $msg = sprintf(shift, @_);
	
	for (split(/\n|(.{425})/, $msg)) {
		raw($sender, 'PRIVMSG %s :%s', $target, $_) if $_;
	}

	$client{$sender}{idle} = time;
}

# notice($sender, $target, $msg, @args)
sub notice {
	my $sender = shift;
	my $target = shift;
	my $msg = sprintf(shift, @_);

	for (split(/\n|(.{425})/, $msg)) {
		raw($sender, 'NOTICE %s :%s', $target, $_) if $_;
	}
}

# operwall($sender, $msg, @args)
sub operwall {
	my $sender = shift;
	my $msg = sprintf(shift, @_);

	raw($sender, 'OPERWALL :%s', $msg);
}

# join_($uid, $chan, $modes, $status)
sub join_ {
	my ($uid, $chan, $modes, $status) = @_;

	if ($chan{$lc{lc($chan)}}) {
		$chan = $lc{lc($chan)};

		raw($uid, 'JOIN %d %s %s', $chan{$chan}{ts}, $chan, $modes);
		if ($status) {
			(my $smodes = $status) =~ y/+%@&~/vhoaq/;
			mode($uid, $chan, "+$smodes $uid");
		}
		$chan{$chan}{users}{$uid} = $status;
		$client{$uid}{chans}{$chan} = $status;
	}
	else {
		raw($conf{sid}, 'SJOIN %d %s %s :%s%s', time, $chan, $modes, $status, $uid);
		callhooks('on_sjoin', $conf{sid}, $chan, time, $modes, $status.$uid);
	}
}

# part($uid, $chan, [$msg])
sub part {
	my ($uid, $chan, $msg) = @_;

	raw($uid, 'PART %s%s', $chan, ($msg ? " :$msg" : ''));
	callhooks('on_part', $uid, $chan, $msg);
}

# quit($uid, [$msg])
sub quit {
	my ($uid, $msg) = @_;

	raw($uid, 'QUIT :%s', $msg);
	callhooks('on_quit', $uid, $msg);
}

# nick($uid, $newnick)
sub nick {
	my ($uid, $newnick) = @_;

	my $time = time;
	raw($uid, 'NICK %s %d', $newnick, $time);
	callhooks('on_nick', $uid, $newnick, $time);
}

# mode($id, $chan, $modes)
sub mode {
	my ($id, $chan, $modes) = @_;
	$chan = $lc{lc($chan)};

	raw($id, 'TMODE %d %s %s', $chan{$chan}{ts}, $chan, $modes);
	callhooks('on_tmode', $id, $chan{$chan}{ts}, $chan, $modes);
}

# kick($kicker, $chan, $target, [msg])
sub kick {
	my ($kicker, $chan, $target, $msg) = @_;

	raw($kicker, 'KICK %s %s :%s', $chan, $target, $msg);
	callhooks('on_kick', $kicker, $chan, $target, $msg);
}

# kill($uid, $msg)
sub kill {
	my ($uid, $msg) = @_;
	$msg ||= 'No reason';

	raw($conf{sid}, 'KILL %s : (%s)', $uid, $msg);
	callhooks('on_quit', $uid, $msg);
}



# svsjoin($nick, $chan)
sub svsjoin {
	my ($nick, $chan) = @_;
	$nick = $lc{lc($nick)};
	$chan = $lc{lc($chan)} if $lc{lc($chan)};

	raw($conf{sid}, 'ENCAP * SVSJOIN %s %s', $nick, $chan);
}

# svspart($nick, $chan) {
sub svspart {
	my ($nick, $chan) = @_;
	$nick = $lc{lc($nick)};
	$chan = $lc{lc($chan)};

	raw($conf{sid}, 'ENCAP * SVSPART %s %s', $nick, $chan);
}

# svsmode($target, $modes)
sub svsmode {
	my ($target, $modes) = @_;
	$target = $lc{lc($target)};

	raw($conf{sid}, 'ENCAP * SVSMODE %s %d :%s', $target, $client{getid($target)}{ts}, $modes);
	callhooks('on_mode', $conf{sid}, getid($target), $modes);
}

# svsnick($nick, $newnick)
sub svsnick {
	my ($nick, $newnick) = @_;
	$nick = $lc{lc($nick)};

	raw($conf{sid}, 'ENCAP * SVSNICK %s %d %s %d', $nick, $client{getid($nick)}{ts}, $newnick, time);
}

# authflags($nick, $authflags)
sub authflags {
	my ($nick, $authflags) = @_;
	$nick = $lc{lc($nick)};

	raw($conf{sid}, 'ENCAP * AUTHFLAGS %s :%s', $nick, $authflags);
	$client{getid($nick)}{authflags} .= $authflags;
}

# capture($uid)
sub capture {
	my $uid = shift;

	raw($conf{sid}, 'CAPTURE :%s', $uid);
}

# uncapture($uid)
sub uncapture {
	my $uid = shift;

	raw($conf{sid}, 'UNCAPTURE :%s', $uid);
}




# on_uid($sid, $nick, $hops, $ts, $umode, $ident, $cloak, $ip, $uid, $host, $gecos)
sub on_uid {
	my ($sid, $nick, $hops, $ts, $umode, $ident, $cloak, $ip, $uid, $host, $gecos) = @_;

	$lc{lc($nick)} = $nick;
	$uid{$nick} = $uid;
	$client{$uid} = {
		sid		=> $sid,
		nick	  => $nick,
		hops	  => $hops,
		ts		 => $ts,
		ident	 => $ident,
		cloak	 => $cloak,
		ip		 => $ip,
		host	  => $host,
		gecos	 => $gecos
	};

	$client{$uid}{umode}{$_} = 1 for split(//, substr($umode, 1));
}

# on_squit($sid, $msg)
sub on_squit {
	my ($sid, $msg) = @_;

	for (keys(%client)) {
		callhooks('on_quit', $_, '*.net *.split') if $client{$_}{sid} eq $sid;
	}

	if ($map{$sid}) {
		on_squit($_) for @{$map{$sid}};
	}

	delete $map{$sid};
	delete $server{$sid};
}

# on_sjoin($sid, $chan, $ts, $modes, @users)
sub on_sjoin {
	my ($sid, $chan, $ts, $modes, @users) = @_;

	if (!$lc{lc($chan)}) {
		$lc{lc($chan)} = $chan;
		$chan{$chan}{ts} = $ts;

		my ($m, @p) = split(/ /, $modes);
		for (split(//, substr($m, 1))) {
			$chan{$chan}{mode}{$_} = ($_ eq 'k' || $_ eq 'l' ? shift(@p) : 1);
		}
	}

	for (@users) {
	my ($uid, $status) = (substr($_, -9), substr($_, 0, length($_) - 9));

		if (!$client{$uid}) {
			debug("OOPS: on_sjoin $chan on non-existent client: $uid");
			next;
		}

		$chan{$chan}{users}{$uid} = $status;
		$client{$uid}{chans}{$chan} = $status;
		callhooks('on_join', $uid, $ts, $chan, $status);
	}
}

# on_part($uid, $chan, $msg)
sub on_part {
	my ($uid, $chan, $msg) = @_;

	delete $chan{$chan}{users}{$uid};
	if (!keys(%{$chan{$chan}{users}}) && !$chan{$chan}{mode}{z}) {
		delete $chan{$chan};
		delete $lc{lc($chan)};
	}
	delete $client{$uid}{chans}{$chan};
}

# on_tmode($id, $ts, $chan, $modes)
sub on_tmode {
	my ($id, $ts, $chan, $modes) = @_;

	my $setby = (length($id) == 3 ? $server{$id} : "$client{$id}{nick}!$client{$id}{ident}\@$client{$id}{cloak}");

	my ($m, @p) = split(/ /, $modes);
	my $add = 1;

	for (split(//, $m)) {
		if ($_ eq '+') {
			$add = 1;
		}
		elsif ($_ eq '-') {
			$add = 0;
		}
		elsif (/([vhoaq])/) {
			(my $prefix = $1) =~ y/vhoaq/+%@&~/;
			my $uid = getid(shift(@p));

			if (!$uid) {
				debug("OOPS: on_tmode $chan $modes on non-existent client");
				next;
			}

			my $newprefix;
			if ($add) {
				$newprefix = join('', sort { $prefix{$a} <=> $prefix{$b} } (split(//, $chan{$chan}{users}{$uid}), $prefix));
			}
			else {
				$newprefix = $chan{$chan}{users}{$uid};
				$newprefix =~ s/\Q$prefix\E//;
			}
			$chan{$chan}{users}{$uid} = $newprefix;
			$client{$uid}{chans}{$chan} = $newprefix;
		}
		elsif ($_ eq 'b') {
			my $ban = shift(@p);
			if ($add) {
				$chan{$chan}{ban}{$ban}{ts} = time;
				$chan{$chan}{ban}{$ban}{setby} = $setby;
			}
			else {
				delete $chan{$chan}{ban}{$ban};
			}
		}
		elsif ($_ eq 'e') {
			my $except = shift(@p);
			if ($add) {
				$chan{$chan}{except}{$except}{ts} = time;
				$chan{$chan}{except}{$except}{setby} = $setby;
			}
			else {
				$chan{$chan}{except}{$except};
			}
		}
		elsif ($_ eq 'I') {
			my $invex = shift(@p);
			if ($add) {
				$chan{$chan}{invex}{$invex}{ts} = time;
				$chan{$chan}{invex}{$invex}{setby} = $setby;
			}
			else {
				delete $chan{$chan}{invex}{$invex};
			}
		}
		elsif ($_ eq 'k' || $_ eq 'l') {
			if ($add) {
				$chan{$chan}{mode}{$_} = shift(@p);
			}
			else {
				delete $chan{$chan}{mode}{$_};
				shift @p if $_ eq 'k';
			}
		}
		else {
			if ($add) {
				$chan{$chan}{mode}{$_} = 1;
			}
			else {
				delete $chan{$chan}{mode}{$_};
				if ($_ eq 'z' && !keys(%{$chan{$chan}{users}})) {
					delete $chan{$chan};
					delete $lc{lc($chan)};
				}
			}
		}
	}
}

# on_mode($id, $target, $modes)
sub on_mode {
	my ($id, $target, $modes) = @_;

	my $add = 1;
	for (split(//, $modes)) {
		next if $_ eq 'd';
		if ($_ eq '+') {
			$add = 1;
		}
		elsif ($_ eq '-') {
			$add = 0;
		}
		else {
			if ($add) {
				$client{$target}{umode}{$_} = 1;
			}
			else {
				delete $client{$target}{umode}{$_};
			}
		}
	}
}

# on_kick($id, $chan, $target, $msg)
sub on_kick {
	my ($id, $chan, $target, $msg) = @_;

	delete $chan{$chan}{users}{$target};
	if (!keys(%{$chan{$chan}{users}}) && !$chan{$chan}{mode}{z}) {
		delete $chan{$chan};
		delete $lc{lc($chan)};
	}
	delete $client{$target}{chans}{$chan};
}

# on_nick($uid, $newnick, $ts)
sub on_nick {
	my ($uid, $newnick, $ts) = @_;
	my $oldnick = $client{$uid}{nick};
	
	#push @{$whowas{lc($oldnick)}}, $client{$uid};
	push @{$client{$uid}{prevnicks}}, $oldnick;
	delete $uid{$oldnick};
	delete $lc{lc($oldnick)};
	$client{$uid}{nick} = $newnick;
	$client{$uid}{ts} = $ts;
	delete $client{$uid}{umode}{r} if $client{$uid}{umode}{r};
	$uid{$newnick} = $uid;
	$lc{lc($newnick)} = $newnick;
}

# on_quit($uid, $msg)
sub on_quit {
	my ($uid, $msg) = @_;

	#push @{$whowas{lc($client{$uid}{nick})}}, $client{$uid};

	for (keys(%{$client{$uid}{chans}})) {
		delete $chan{$_}{users}{$uid};
		if (!keys(%{$chan{$_}{users}}) && !$chan{$_}{mode}{z}) {
			delete $chan{$_};
			delete $lc{lc($_)};
		}
	}
	delete $uid{$client{$uid}{nick}};
	delete $lc{lc($client{$uid}{nick})};
	delete $client{$uid};
}
